"""Вспомогательная структура для хранения списка доступных команд"""
from . import baseCommands
from .badSituations import NotImplementedCommand


class CommandList(dict):
    @staticmethod
    def get_command_by_uuid(module, uuid):
        """Из определённого модуля получает имя команды по UUID"""
        for cls in dir(module):
            obj = getattr(module, cls)
            if hasattr(obj, 'COMMAND_UUID') and getattr(obj, 'COMMAND_UUID') == uuid:
                return obj
        raise NotImplementedCommand(f'command {uuid} not implemented')

    def __init__(self, module, module_impl):
        """В конструктор передаётся модуль со списком команд и модуль со списком реализаций"""
        super().__init__(self)
        for field in dir(module):
            obj = getattr(module, field)
            if isinstance(obj, baseCommands.REGISTRY_COMMAND):
                uuid = obj.uuid
                # 0 - CommandName
                # 1 - CommandUUID
                # 2 - CommandRealisation
                self[uuid] = (obj.name, uuid, CommandList.get_command_by_uuid(module_impl, uuid))

    def get_command_impl(self, command_uuid):
        """По UUID команды получить её реализацию"""
        return self[command_uuid][2]

    def get_command_name(self, command_uuid):
        """По UUID команды получаем Имя команды"""
        if command_uuid in self:
            return self[command_uuid][0]
        return None



import socket
from threading import Thread

from .tcpWorker import TcpWorker
from .connection import Connection
from .logger import logger


class TcpServer(TcpWorker):
    """
    Сущность, которая выступает в роли слушателя
    к ней можно подключится образовав тем самым конекцию, для дальнейшего обмена сообщениями
    """
    def __init__(self, ip, port, client_commands, client_command_impl):
        super().__init__(ip, port, client_commands, client_command_impl)

        self.serv_socket = socket.socket()

    def connect_listener(self, new_client_handler):
        while True:
            sock, adr = self.serv_socket.accept()
            connection = Connection(self, sock)
            logger.info(f'{connection.getpeername()} - was connected')
            self.run_connection(connection)

            if new_client_handler:
                new_client_handler(connection)

    def run(self, new_connection_handler=None, disconect_connection_handler=None):
        self.serv_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.serv_socket.bind((self.ip, self.port))
        self.serv_socket.listen(10)

        thread = Thread(target=self.connect_listener, args=(new_connection_handler,))
        thread.daemon = True
        thread.start()

        self.set_disconnection_handler(disconect_connection_handler)

    def stop(self):
        self.finish_all(0, 'Good bye!')
        self.serv_socket.shutdown(socket.SHUT_RDWR)
        self.serv_socket.close()



"""Список констант с базовыми командами"""
from .badSituations import MessageStatusFailed, MessageStatusError
from .message import Message
from .logger import logger


Unknown = 'UNKNOWN'
Error = 'ERROR'
ProtocolCompatible = 'PROTOCOL_COMPATIBLE'
CloseConnection = 'CLOSE_CONNECTION'


class REGISTRY_COMMAND:
    commands_names = []
    commands_uuids = []

    def __init__(self, name, uuid):
        self.name = name
        self.uuid = uuid
        if name in REGISTRY_COMMAND.commands_names:
            raise ValueError(f'Command name {name} is not unique')
        if uuid in REGISTRY_COMMAND.commands_uuids:
            raise ValueError(f'Command uuid {uuid} is not unique')
        REGISTRY_COMMAND.commands_names.append(name)
        REGISTRY_COMMAND.commands_uuids.append(uuid)


# Регистрация базовых команд
UNKNOWN             = REGISTRY_COMMAND(Unknown,            "4aef29d6-5b1a-4323-8655-ef0d4f1bb79d")
ERROR               = REGISTRY_COMMAND(Error,              "b18b98cc-b026-4bfe-8e33-e7afebfbe78b")
PROTOCOL_COMPATIBLE = REGISTRY_COMMAND(ProtocolCompatible, "173cbbeb-1d81-4e01-bf3c-5d06f9c878c3")
CLOSE_CONNECTION    = REGISTRY_COMMAND(CloseConnection,    "e71921fd-e5b3-4f9b-8be7-283e8bb2a531")


class BaseCommand:
    """
    Шаблон реализации всех команд
    Чтобы зарегистрировать в воркере свою команду, необходимо наследоваться от этого класса и
    реализовать необходимые команды.
    Пример реализации расположен в файле baseCommandsImpl, где с помощью этого класса реализованы
    базовые команды протокола
    """
    COMMAND_UUID = None

    @staticmethod
    def initial(connection, *args, **kwargs):
        """
        Метод вызывается перед отправкой любой команды, он должен вернуть сформированный message
        для последующей отправки
        """
        raise NotImplemented('Initialization method not implemented yet')

    @staticmethod
    def answer(msg: Message):
        """
        Метод обработчик, срабатывает в случае, когда на команду приходит ответ, с тем же идентификатором.
        Сюда мы можем попасть с сообщение типа Command и Answer
        """
        pass

    @staticmethod
    def answer_fail(msg: Message):
        """
        Метод обработчик ответа на команду, вызывается в том случает когда отет приходит со статусом ExecStatus.Failed
        """
        raise MessageStatusFailed(msg, f'Command {msg.get_id()} was failed')

    @staticmethod
    def answer_error(msg: Message):
        """
        Метод обработчик ответа на команду, вызывается в том случает когда отет приходит со статусом ExecStatus.Failed
        """
        raise MessageStatusError(msg, f'Command {msg.get_id()} was error')

    @staticmethod
    def handler(msg: Message):
        """
        Метод обработчик входящей команды, идентификатор которой не найден в списке запросов.
        Скорее всего это значит что вторая сторона, отправила команду, но также сюда можно попасть по какой-либо ошибке.
        Так же как и в обработчике answer сюда можно попасть с типом и Command и Answer
        """
        raise Exception('Message processing method not implemented yet')

    @staticmethod
    def handler_sync(msg: Message):
        raise Exception('Message sync handler is not implemented')

    @staticmethod
    def unknown(msg: Message):
        """
        Обработчик ситуации, при которой в ответ на команду приходит сообщение о том, что данная команда неизвестна
        """
        logger.info(f'[{msg.my_connection.getpeername()}] Unknown command for remote client! {msg.get_id()}')
        # raise Exception('Команда неизвестна для удалённого клиента!')

    @staticmethod
    def timeout(msg: Message):
        """
        Если в сообщении задать максимальное время выполнения команды,
        в случае истечения времени сработает этот обработчик
        """
        raise TimeoutError('Waiting command execution timed out')

    @classmethod
    def exec_decorator(cls, connection):
        def function_template(*args, **kwargs):
            connection.exec_command(cls, *args, **kwargs)
        return function_template

    @classmethod
    def sync_decorator(cls, connection):
        def function_template(*args, **kwargs):
            return connection.exec_command_sync(cls, *args, **kwargs)
        return function_template

    @classmethod
    def async_decorator(cls, connection):
        def function_template(*args, **kwargs):
            connection.exec_command_async(cls, *args, **kwargs)
        return function_template

    @classmethod
    def sync_handler_decorator(cls, connection):
        def function_template():
            return connection.catch_handler(cls)
        return function_template



"""Базовый init файл, с импортом наиболее используемых объектов из всей библиотеки proto_py"""


from .tcpSocket import TcpSocket
from .tcpServer import TcpServer
from .message import Message
from .connection import Connection
from .baseCommands import BaseCommand, REGISTRY_COMMAND



import json
from copy import copy
from uuid import uuid4

from .badSituations import NotConnectionException, UnknownCommandRecieved
from .logger import logger
from .tools import try_uuid
from .flags import MsgFlag, Type, ExecStatus


class Message(dict):
    """
    Объект, преобразования которого отправляется по сети.
    Он хранит в себе всю необходимую информаци о пакете который мы хотим отправить по сети:
    ID
    Тип
    Команду
    Контент
    Тэги
    Флаги
    Мин/Макс время жизни
    """
    TCP_FIELDS = ['id', 'command', 'flags', 'content', 'PROTOCOL_VERSION_LOW',
                  'PROTOCOL_VERSION_HIGH', 'tags', 'maxTimeLife']

    def __str__(self):
        """Метод для преобразования в строку для записи объекта в логах"""
        res = []

        # Определимся с типом сообщения
        type_name = Type.Unknown
        type_value = self.get_type()
        if type_value == Type.Command:
            type_name = 'Command'
        elif type_value == Type.Answer:
            type_name = 'Answer'
        elif type_value == Type.Event:
            type_name = 'Event'
        res.append(f'|Type: {type_name}|')

        for field in self:
            if field in ['command']:
                continue
            res.append(f'{field}: {self[field]}')
        return ', '.join(res)

    def to_serializable_dict(self):
        temp_json = dict(self)
        temp_json['flags'] = str(temp_json.get('flags', ''))
        return temp_json

    def __init__(self, connection, id_=None, command_uuid=None):
        self.my_worker = connection.worker
        self.my_connection = connection
        # ID или создаётся, если это инициализация нового сообщения, или задаётся, если это парсинг сообщения из сети
        if id_ is not None:
            if try_uuid(id_):
                self['id'] = id_
            else:
                raise ValueError('Invalid packet identifier format')
        else:
            self['id'] = str(uuid4())

        # если при создании передан uuid команды, сразу проверим зарегистрирована ли такая команда,
        # и пропишем всю информацию в объекте
        if command_uuid is not None:
            comm_name = self.my_worker.get_command_name(command_uuid)
            if comm_name is not None:
                self['command'] = command_uuid
                self['Command'] = comm_name
            else:
                raise UnknownCommandRecieved(f'Unknown command uuid {command_uuid}')

        # месседж не может существовать без флагов, при создании они инициализируются дефолтными значениями
        self['flags'] = MsgFlag()

    # Список сеттеров и геттеров, для более удобного обращения со стороны пользователя
    # ===================================================================================
    def set_connection(self, conn):
        self.my_connection = conn

    def get_answer_copy(self):
        """Из входящего сообщения, создаёт копию, устанавливает тип ответ и статус Success"""
        answer = copy(self)
        answer.set_type(Type.Answer)
        answer.set_status(ExecStatus.Success)
        return answer

    def set_type(self, type_):
        self['flags'].set_flag_value('type', type_)

    def get_type(self):
        return self['flags'].get_flag_value('type')

    def set_status(self, status):
        self['flags'].set_flag_value('execStatus', status)

    def get_status(self):
        return self['flags'].get_flag_value('execStatus')

    def set_content(self, content):
        self['content'] = content
        self['flags'].set_flag_value('contentIsEmpty', 0 if content else 1)

    def set_protocol_version_low(self, version):
        self['PROTOCOL_VERSION_LOW'] = version

    def set_protocol_version_high(self, version):
        self['PROTOCOL_VERSION_HIGH'] = version

    def get_id(self):
        return self['id']

    def get_command(self):
        """Возвращает str(UUID) команды сообщения"""
        return self['command']

    def get_content(self):
        return self.get('content')

    def set_tag(self, value, num):
        tags = self.get('tags')
        if not tags:
            tags = [0] * (num+1)
            tags[num] = value
            self['tags'] = tags
            return

        if len(tags) > num:
            tags[num] = value
        else:
            tags += [0]*(num - len(tags) + 1)
            tags[num] = value

    def tag(self, num=0):
        if num not in range(0, 254):
            logger.error('Index value not in range [0..254]')
            return 0

        if 'tags' not in self or len(self.get('tags')) < num + 1:
            return 0

        return self.get('tags')[num]

    def set_max_time_life(self, max_time_life):
        """Устанавлдивает время ожидания в секундах"""
        self['maxTimeLife'] = max_time_life

    def get_max_time_life(self):
        return self.get('maxTimeLife')

    def get_bytes(self):
        result = dict()
        for f in Message.TCP_FIELDS:
            if f is None:
                continue
            if f == 'flags':
                result[f] = self[f].get_digit()
            elif f in self:
                result[f] = self[f]
        return json.dumps(result).encode()
    # Конец списка сеттеров и геттеров
    # ====================================================================================

    def send_answer(self):
        """
        В случае если нужно отправить ответ на команду, и у нас есть все данные и о воркере и о конеции
        можно использовать метод send_message от объекта Message
        В ином случае, нужно отправлять сообщение используя метод Connection::send_message
        """
        if self.my_connection is None:
            raise NotConnectionException('отсутсвует коннекция, отправка невозможна')

        if self.get_type() == Type.Answer:
            self.my_connection.send_message(self)

    # список статических методов для создания наиболее популярных типов сообщений
    # с пресетами свойств(флагов, типов, контентов и т.п.
    @staticmethod
    def command(connection, command_uuid):
        """Статический метод создания сообщения с типом команды"""
        msg = Message(connection, command_uuid=command_uuid)
        msg.set_type(Type.Command)
        return msg

    @staticmethod
    def event(connection, command_uuid):
        """Статический метод создания сообщения с типом команды"""
        msg = Message(connection, command_uuid=command_uuid)
        msg.set_type(Type.Event)
        return msg

    @staticmethod
    def answer(connection, command_uuid):
        """Статический метод создания сообщения с типом ответ"""
        msg = Message(connection, command_uuid=command_uuid)
        msg.set_type(Type.Answer)
        return msg
    # Конец списка



"""Список namespace'ов для описания существующих типов флагов"""


class SocketType:
    Unknown = 0
    Local = 1
    Tcp = 2
    Udp = 3


class SerializationFormat:
    BProto = 0      # Бинарный формат
    Json = 1        # Json формат
    # LastFormat = 7 Предполагается, что будет не больше 8 форматов


class Type:
    Unknown = 0
    Command = 1     # Сообщение-команда. Это может быть сообщение с командой
                    # на выполнение действия, либо это может быть запрос
                    # на получение данных. Предполагается, что в ответ
                    # на данное сообщение придет сообщение с типом Answer.
    Answer = 2      # Ответ на сообщением с типом Command.
    Event = 3       # Данный тип сообщения похож на Command, но не предполагает
                    # получения ответа (Answer). Он используется для рассылки
                    # широковещательных сообщений о событиях


class ExecStatus:
    Unknown = 0
    Success = 1     # Сообщение было обработано успешно и содержит корректные
                    # ответные данные.
    Failed = 2      # Сообщение не было обработано успешно, но результат
                    # не является ошибкой.
                    # В данном случае сообщение (Message) будет содержать
                    # данные в формате communication::data::MessageFailed.
    Error = 3       # При обработке сообщения произошла ошибка, и в качестве
                    # ответа отправляется сообщения с описанием причины ошибки.
                    # В данном случае сообщение (Message) будет содержать
                    # данные в формате communication::data::MessageError


class Priority:
    High = 0
    Normal = 1
    Low = 2
    # Reserved = 3


class Compression:
    NoneCompression = 0
    Zip = 1
    Lzma = 2
    Ppmd = 3
    Disable = 7     # Используется в тех случаях когда нужно явно запретить
                    # сжатие сообщения при его отправке в TCP сокет.
                    # Это может потребоваться когда контент изначально сжат,
                    # например, при отправке JPG, PNG, и прочих подобных
                    # форматов


class FlagField:
    """Вспомогательная структура для хранения информации о конкретном флаге"""
    def __init__(self, name, size, value):
        self.name = name
        self.size = size
        self.value = value


class MsgFlag:
    """Структура для хранения информации о всех флагах сообщения"""
    def __str__(self):
        return str(self.get_digit())

    def __init__(self):
        self.values = [
            # 1 байт
            FlagField('type', 3, Type.Unknown),
            FlagField('execStatus', 3, ExecStatus.Unknown),
            FlagField('priority', 2, Priority.Normal),

            # 2 байт
            FlagField('compression', 3, Compression.Disable),
            FlagField('tagsIsEmpty', 1, 0),
            FlagField('maxTimeLifeIsEmpty', 1, 0),
            FlagField('contentIsEmpty', 1, 1),
            FlagField('reserved2', 2, 0),

            # 3 байт
            FlagField('reserved3', 8, 0),

            # 4 байт
            FlagField('contentFormat', 3, SerializationFormat.Json),
            FlagField('reserved4', 4, 0),
            FlagField('flags2IsEmpty', 1, 0),
        ]

    def set_flag_value(self, name, value):
        """Установить значение конкретного флага, принимает имя и значение, проверяет допустимую величину для флага"""
        for f in self.values:
            if f.name == name:
                if 2 ** f.size <= value:
                    raise ValueError('Превышено допустимое значение для флага')
                f.value = value

    def get_flag_value(self, name) -> int:
        """Получение значения флага, по имени флага"""
        for f in self.values:
            if f.name == name:
                return f.value

    def get_digit(self) -> int:
        """Преобразует значение структуры хранения данных о флаге в число, которое потом можно конвертировать
        в байтовое представление и отправить по сети"""
        res = ''
        for field in reversed(self.values):
            # число значения флага превращает в бинарное представление с ведущими нулями
            res += bin(field.value)[2:].zfill(field.size)
        # из строки вида "01010101" получаем число
        return int(res, 2)

    @staticmethod
    def from_digit(digit):
        """Из числа преобразует в структуру хранения данных флагов
        операция обратная операции get_digit"""
        b_string = bin(digit)[2:].zfill(8 * 4)
        flag = MsgFlag()
        for f in reversed(flag.values):
            f.value = int(b_string[0:f.size], 2)
            b_string = b_string[f.size:]
        return flag



from setuptools import setup, find_packages


# with open("README.md", "r") as fh:
#     long_description = fh.read()

setup(
    name='pproto-py',
    version='1.1.9',
    description='pproto_py is Python implementation of "Point Of View" communication protocol',
    url="https://github.com/TochkaAI/pproto_py",
    packages=find_packages(),
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from threading import Thread

from .badSituations import UnknownCommandRecieved
from .commandList import CommandList
from .connection import Connection
from .logger import logger
from . import baseCommands
from . import baseCommandsImpl
from .baseCommandsImpl import CloseConnectionCommand, UnknownCommand
from .connectionPool import ConnectionPool


class TcpWorker:
    """
    Главная сущность осуществляющая основную логику общения между клиентами в рамках оговоренного протокола
    """
    user_commands_list: CommandList
    base_commands_list: CommandList

    def __init__(self, ip, port, client_commands, client_command_impl):
        self.ip = ip
        self.port = port

        self.connection_pool = ConnectionPool()
        self.disconnection_handler = None

        self.base_commands_list = CommandList(baseCommands, baseCommandsImpl)
        self.user_commands_list = CommandList(client_commands, client_command_impl)

        self.unknown_command_list = []

    def set_disconnection_handler(self, handler):
        """Если пользователь пожелает задать обработчик на разрыв соединения, то присвоим его значение тут"""
        self.disconnection_handler = handler

    def _cmd_method_creator(self, connection):
        """Создаёт у коннекции всевозможные доступные пользовательские и базовые методы"""
        if connection.user_command_recorded:
            return
        connection.user_command_recorded = True
        for cmd in list(self.user_commands_list.values()) + list(self.base_commands_list.values()):
            setattr(connection, cmd[0] + '_exec', cmd[2].exec_decorator(connection))
            setattr(connection, cmd[0] + '_sync', cmd[2].sync_decorator(connection))
            setattr(connection, cmd[0] + '_async', cmd[2].async_decorator(connection))
            setattr(connection, cmd[0] + '_catch', cmd[2].sync_handler_decorator(connection))

    def get_command_name(self, command_uuid):
        """По UUID команды получаем Имя команды из списка базовых или из списка пользовательских"""
        return (self.base_commands_list.get_command_name(command_uuid) or
                self.user_commands_list.get_command_name(command_uuid))

    def command_handler(self, msg):
        if msg.get_command() in self.base_commands_list:
            # Обработчики базовых команд
            command = self.base_commands_list.get_command_impl(msg.get_command())
        else:
            # Обработчики пользовательских команд
            command = self.user_commands_list.get_command_impl(msg.get_command())
        # и теперь отправим обрабатывать это в отдельный поток, чтоб не тормозило получение новых команд
        thread = Thread(target=command.handler, args=(msg,))
        thread.daemon = True
        thread.start()

    def start_listening(self, connection):
        """Эта команда для конекции запускает бесконечный цикл прослушивания сокета"""
        thread = Thread(target=self.command_listener, args=(connection,))
        thread.daemon = True
        thread.start()

    def command_listener(self, connection: Connection):
        """Метод запускается в отдельном потоке и мониторит входящие пакеты
        пытается их распарсить и выполнить их соответствующую обработку"""
        while True:
            json_data = connection.mrecv()
            if json_data:
                logger.info(f'[{connection.getpeername()}] JSON received: {json_data}')
                try:
                    msg = connection.message_from_json(json_data)  # Type: Message
                except UnknownCommandRecieved:
                    logger.info(f'[{connection.getpeername()}] Unknown msg received')
                    connection.exec_command(UnknownCommand, json_data)
                else:
                    logger.info(f'[{connection.getpeername()}] Msg  received: {msg}')
                    # Это ответы, которые нужно обработать, в синхронном или асинхронном режиме
                    if msg.get_id() in connection.request_pool or \
                            connection.sync_handler_pool.is_catching(msg.get_command()):
                        connection.message_pool.add_message(msg)
                    # Это команда с той стороны, её нужно передать в соответсвующий handler
                    else:
                        self.command_handler(msg)
            else:
                break
        # соответственно если я тут, значит у нас произошёл разрыв соединения
        self.connection_pool.del_connection(connection)
        if connection.is_connected():  # тут нужна проверка, потому что мы сами могли порвать соединение
            connection.close()

    def run_connection(self, connection: Connection):
        """Функция которая выполняет стандартный сценарий, сразу после образования Tcp соединения"""
        self._cmd_method_creator(connection)
        connection.start()
        self.connection_pool.add_connection(connection)
        self.start_listening(connection)

    def finish_all(self, code, description):
        """Функция завершает все соединения предварительно отправив команду CloseConnection"""
        for conn in list(self.connection_pool.values()):
            peer_name = conn.getpeername()
            conn.exec_command_sync(CloseConnectionCommand, code, description)
            conn.close()
            logger.info(f'[{peer_name}] Disconect from host')



"""Файл содержащий реализации исключений для всевозможных нештатных ситуаций"""


class ProtocolIncompatibleEx(BaseException):
    def __init__(self, msg):
        super().__init__(msg)


class UnknownCommandSend(BaseException):
    def __init__(self, msg):
        super().__init__(msg)


class UnknownCommandRecieved(BaseException):
    def __init__(self, msg):
        super().__init__(msg)


class NotImplementedCommand(BaseException):
    def __init__(self, msg):
        super().__init__(msg)


class NotConnectionException(BaseException):
    def __init__(self, msg):
        super().__init__(msg)


class MessageStatusFailed(BaseException):
    def __init__(self, msg, info):
        self.failed_msg = msg
        super().__init__(info)


class MessageStatusError(BaseException):
    def __init__(self, msg, info):
        self.error_msg = msg
        super().__init__(info)
